@@ -25,7 +25,7 @@ fn main() {
             format!("bin/{}.a", target),
             out_dir.join(format!("lib{}.a", name)),
         )
-        .unwrap();
+        .expect("couldn't find asm support library for target platform");
 
         println!("cargo:rustc-link-lib=static={}", name);
         println!("cargo:rustc-link-search={}", out_dir.display());
