@@ -11,6 +11,52 @@ pub type Connection = usize;
 
 pub const MAX_CID: usize = 34;
 
+pub const FLASH_PHYS_BASE: u32 = 0x2000_0000;
+pub const SOC_REGION_LOC: u32 = 0x0000_0000;
+pub const SOC_REGION_LEN: u32 = 0x00D0_0000; // gw + staging + loader + kernel
+
+pub const SOC_MAIN_GW_LOC: u32 = 0x0000_0000; // gateware - primary loading address
+pub const SOC_MAIN_GW_LEN: u32 = 0x0028_0000;
+pub const SOC_STAGING_GW_LOC: u32 = 0x0028_0000; // gateware - staging copy
+pub const SOC_STAGING_GW_LEN: u32 = 0x0028_0000;
+
+pub const LOADER_LOC: u32 = 0x0050_0000; // loader - base
+pub const LOADER_CODE_LEN: u32 = 0x0003_0000; // code region only
+pub const LOADER_FONT_LOC: u32 = 0x0053_0000; // should be the same as graphics-server/src/fontmap.rs/FONT_BASE
+pub const LOADER_FONT_LEN: u32 = 0x0044_0000; // length of font region only
+pub const LOADER_TOTAL_LEN: u32 = LOADER_CODE_LEN + LOADER_FONT_LEN; // code + font
+
+pub const EARLY_SETTINGS: u32 = 0x0097_0000;
+
+pub const KERNEL_LOC: u32 = 0x0098_0000; // kernel start
+pub const KERNEL_LEN: u32 = 0x0140_0000; // max kernel length = 0xA0_0000 * 2 => half the area for backup kernel & updates
+
+pub const EC_REGION_LOC: u32 = 0x07F8_0000; // EC update staging area. Must be aligned to a 64k-address.
+pub const EC_WF200_PKG_LOC: u32 = 0x07F8_0000;
+pub const EC_WF200_PKG_LEN: u32 = 0x0004_E000;
+pub const EC_FW_PKG_LOC: u32 = 0x07FC_E000;
+pub const EC_FW_PKG_LEN: u32 = 0x0003_2000;
+pub const EC_REGION_LEN: u32 = 0x0008_0000;
+
+pub const PDDB_LOC: u32 = 0x01D8_0000; // PDDB start
+pub const PDDB_LEN: u32 = EC_REGION_LOC - PDDB_LOC; // must be 64k-aligned (bulk erase block size) for proper function.
+
+#[cfg(not(any(target_os = "none", target_os = "xous")))]
+use core::sync::atomic::AtomicU64;
+
+
+// Secretly, you can change this by setting the XOUS_SEED environment variable.
+// I don't lke environment variables because where do you document features like this?
+// But, this was the most expedient way to get all the threads in Hosted mode to pick up a seed.
+// The code that reads the varable this is all the way over in xous-rs\src\arch\hosted\mod.rs#29, and
+// it's glommed onto some other static process initialization code because I don't fully understand
+// what's going on over there.
+#[cfg(not(any(target_os = "none", target_os = "xous")))]
+pub static TESTING_RNG_SEED: AtomicU64 = AtomicU64::new(0);
+
+pub mod exceptions;
+pub use exceptions::*;
+
 #[derive(Debug, PartialEq, Eq, Ord, PartialOrd, Copy, Clone)]
 pub struct MessageSender {
     data: usize,
@@ -44,7 +90,7 @@ impl core::fmt::Display for MessageSender {
 }
 
 /// Server ID
-#[derive(Debug, Copy, Clone, PartialEq)]
+#[derive(Debug, Copy, Clone, PartialEq, Eq)]
 pub struct SID([u32; 4]);
 impl SID {
     pub fn from_bytes(b: &[u8]) -> Option<SID> {
@@ -102,9 +148,9 @@ impl From<&[u32; 4]> for SID {
     }
 }
 
-impl Into<[u32; 4]> for SID {
-    fn into(self) -> [u32; 4] {
-        self.0
+impl From<SID> for [u32; 4] {
+    fn from(s: SID) -> [u32; 4] {
+        s.0
     }
 }
 
@@ -119,8 +165,8 @@ pub type CpuID = usize;
 
 #[derive(Debug, PartialEq, Copy, Clone)]
 pub struct MemoryRange {
-    pub addr: MemoryAddress,
-    pub size: MemorySize,
+    pub(crate) addr: MemoryAddress,
+    pub(crate) size: MemorySize,
 }
 
 #[cfg(feature = "bitflags")]
@@ -171,7 +217,7 @@ pub fn pid_from_usize(src: usize) -> core::result::Result<PID, Error> {
     if src > u8::MAX as _ {
         return Err(Error::InvalidPID);
     }
-    Ok(PID::new(src as u8).ok_or(Error::InvalidPID)?)
+    PID::new(src as u8).ok_or(Error::InvalidPID)
 }
 
 #[repr(usize)]
@@ -203,6 +249,7 @@ pub enum Error {
     AccessDenied = 23,
     UseBeforeInit = 24,
     DoubleFree = 25,
+    DebugInProgress = 26,
 }
 
 impl Error {
@@ -234,6 +281,7 @@ impl Error {
             23 => AccessDenied,
             24 => UseBeforeInit,
             25 => DoubleFree,
+            26 => DebugInProgress,
             _ => UnknownError,
         }
     }
@@ -265,6 +313,7 @@ impl Error {
             AccessDenied => 23,
             UseBeforeInit => 24,
             DoubleFree => 25,
+            DebugInProgress => 26,
             UnknownError => usize::MAX,
         }
     }
@@ -458,7 +507,7 @@ impl Message {
 
     pub fn memory_message(&self) -> Option<&MemoryMessage> {
         match self {
-            Message::MutableBorrow(mem) | Message::Borrow(mem) | Message::Move(mem) => Some(&mem),
+            Message::MutableBorrow(mem) | Message::Borrow(mem) | Message::Move(mem) => Some(mem),
             Message::BlockingScalar(_) | Message::Scalar(_) => None,
         }
     }
@@ -473,7 +522,7 @@ impl Message {
     pub fn scalar_message(&self) -> Option<&ScalarMessage> {
         match self {
             Message::MutableBorrow(_) | Message::Borrow(_) | Message::Move(_) => None,
-            Message::BlockingScalar(scalar) | Message::Scalar(scalar) => Some(&scalar),
+            Message::BlockingScalar(scalar) | Message::Scalar(scalar) => Some(scalar),
         }
     }
 
@@ -504,19 +553,19 @@ impl TryFrom<(usize, usize, usize, usize, usize, usize)> for Message {
         match value.0 {
             1 => Ok(Message::MutableBorrow(MemoryMessage {
                 id: value.1,
-                buf: MemoryRange::new(value.2, value.3).map_err(|_| ())?,
+                buf: unsafe { MemoryRange::new(value.2, value.3).map_err(|_| ()) }?,
                 offset: MemoryAddress::new(value.4),
                 valid: MemorySize::new(value.5),
             })),
             2 => Ok(Message::Borrow(MemoryMessage {
                 id: value.1,
-                buf: MemoryRange::new(value.2, value.3).map_err(|_| ())?,
+                buf: unsafe { MemoryRange::new(value.2, value.3).map_err(|_| ()) }?,
                 offset: MemoryAddress::new(value.4),
                 valid: MemorySize::new(value.5),
             })),
             3 => Ok(Message::Move(MemoryMessage {
                 id: value.1,
-                buf: MemoryRange::new(value.2, value.3).map_err(|_| ())?,
+                buf: unsafe { MemoryRange::new(value.2, value.3).map_err(|_| ()) }?,
                 offset: MemoryAddress::new(value.4),
                 valid: MemorySize::new(value.5),
             })),
@@ -579,6 +628,16 @@ impl Drop for MessageEnvelope {
                     .expect("couldn't return memory")
             }
             Message::Move(msg) => {
+                // Reconstitute the vec and drop it right away to ensure it's freed.
+                #[cfg(any(unix, windows))]
+                let _v = unsafe {
+                    std::vec::Vec::from_raw_parts(
+                        msg.buf.as_mut_ptr(),
+                        msg.buf.len(),
+                        msg.buf.len(),
+                    )
+                };
+                #[cfg(not(any(unix, windows)))]
                 crate::syscall::unmap_memory(msg.buf).expect("couldn't free memory message")
             }
             _ => (),
@@ -587,7 +646,12 @@ impl Drop for MessageEnvelope {
 }
 
 impl MemoryRange {
-    pub fn new(addr: usize, size: usize) -> core::result::Result<MemoryRange, Error> {
+    /// # Safety
+    ///
+    /// This allows for creating a `MemoryRange` from any arbitrary pointer,
+    /// so it is imperitive that this only be used to point to valid, page-aligned
+    /// ranges.
+    pub unsafe fn new(addr: usize, size: usize) -> core::result::Result<MemoryRange, Error> {
         assert!(
             addr != 0,
             "tried to construct a memory range with a null pointer"
@@ -599,6 +663,7 @@ impl MemoryRange {
         })
     }
 
+    #[deprecated(since = "0.8.4", note = "Please use `new(addr, size)` instead")]
     pub fn from_parts(addr: MemoryAddress, size: MemorySize) -> MemoryRange {
         MemoryRange { addr, size }
     }
@@ -618,6 +683,40 @@ impl MemoryRange {
     pub fn as_mut_ptr(&self) -> *mut u8 {
         self.addr.get() as *mut u8
     }
+
+    /// Return this memory as a slice of values. The resulting slice
+    /// will cover the maximum number of elements given the size of `T`.
+    /// For example, if the allocation is 4096 bytes, then the resulting
+    /// `&[u8]` would have 4096 elements, `&[u16]` would have 2048, and
+    /// `&[u32]` would have 1024. Values are rounded down.
+    pub fn as_slice<T>(&self) -> &[T] {
+        // This is safe because the pointer and length are guaranteed to
+        // be valid, as long as the user hasn't already called `as_ptr()`
+        // and done something unsound with the resulting pointer.
+        unsafe {
+            core::slice::from_raw_parts(
+                self.as_ptr() as *const T,
+                self.len() / core::mem::size_of::<T>(),
+            )
+        }
+    }
+
+    /// Return this memory as a slice of mutable values. The resulting slice
+    /// will cover the maximum number of elements given the size of `T`.
+    /// For example, if the allocation is 4096 bytes, then the resulting
+    /// `&[u8]` would have 4096 elements, `&[u16]` would have 2048, and
+    /// `&[u32]` would have 1024. Values are rounded down.
+    pub fn as_slice_mut<T>(&mut self) -> &mut [T] {
+        // This is safe because the pointer and length are guaranteed to
+        // be valid, as long as the user hasn't already called `as_ptr()`
+        // and done something unsound with the resulting pointer.
+        unsafe {
+            core::slice::from_raw_parts_mut(
+                self.as_mut_ptr() as *mut T,
+                self.len() / core::mem::size_of::<T>(),
+            )
+        }
+    }
 }
 
 /// Which memory region the operation should affect.
@@ -698,7 +797,7 @@ pub enum Result {
 
     /// The syscall should be attempted again. This is returned when calling
     /// functions such as `try_connect()` and `try_send()` that may block.
-    WouldBlock,
+    RetryCall,
 
     /// The message was successful but no value was returned.
     None,
@@ -753,7 +852,7 @@ impl Result {
                     0,
                 ]
             }
-            Result::WouldBlock => [16, 0, 0, 0, 0, 0, 0, 0],
+            Result::RetryCall => [16, 0, 0, 0, 0, 0, 0, 0],
             Result::None => [17, 0, 0, 0, 0, 0, 0, 0],
             Result::MemoryReturned(offset, valid) => [
                 18,
@@ -838,7 +937,7 @@ impl Result {
                 SID::from_u32(src[1] as _, src[2] as _, src[3] as _, src[4] as _),
                 src[5] as _,
             ),
-            16 => Result::WouldBlock,
+            16 => Result::RetryCall,
             17 => Result::None,
             18 => Result::MemoryReturned(MemorySize::new(src[1]), MemorySize::new(src[2])),
             _ => Result::UnknownResult(src[0], src[1], src[2], src[3], src[4], src[5], src[6]),
