@@ -4,12 +4,15 @@ use crate::{
     Result, ScalarMessage, SysCallResult, ThreadInit, CID, PID, SID, TID,
 };
 use core::convert::{TryFrom, TryInto};
+/* https://github.com/betrusted-io/xous-core/issues/90
+use crate::Exception
+*/
 
 // use num_derive::FromPrimitive;
 // use num_traits::FromPrimitive;
 
-#[cfg(not(any(target_os = "none", target_os = "xous")))]
-use crate::ProcessArgsAsThread;
+#[cfg(feature = "processes-as-threads")]
+pub use crate::arch::ProcessArgsAsThread;
 
 #[derive(Debug, PartialEq)]
 pub enum SysCall {
@@ -20,6 +23,10 @@ pub enum SysCall {
     /// If a virtual address is specified, then the returned pages are located
     /// at that address.  Otherwise, they are located at the Default offset.
     ///
+    /// # Returns
+    ///
+    /// * **MemoryRange**: A memory range containing zeroed bytes.
+    ///
     /// # Errors
     ///
     /// * **BadAlignment**: Either the physical or virtual addresses aren't
@@ -38,6 +45,9 @@ pub enum SysCall {
     ///
     /// # Errors
     ///
+    /// * **BadAlignment**: The memory range was not page-aligned
+    /// * **BadAddress**: A page in the range was not mapped
+    ///
     UnmapMemory(MemoryRange),
 
     /// Sets the offset and size of a given memory region.  This call may only
@@ -98,6 +108,13 @@ pub enum SysCall {
     /// to REMOVE flags on a memory region, for example to mark it as no-execute
     /// after writing program data.
     ///
+    /// If `PID` is `None`, then modifies this process. Note that it is not legal
+    /// to modify the memory range of another process that has been started already.
+    ///
+    /// # Returns
+    ///
+    /// * **Ok**: The call completed successfully
+    ///
     /// # Errors
     ///
     /// * **ProcessNotChild**: The given PID is not a child of the current
@@ -105,26 +122,65 @@ pub enum SysCall {
     /// * **MemoryInUse**: The given PID has already been started, and it is not
     ///                    legal to modify memory flags anymore.
     UpdateMemoryFlags(
-        MemoryAddress, /* virt */
-        usize,         /* number of pages */
-        MemoryFlags,   /* new flags */
+        MemoryRange, /* range of memory to update flags for */
+        MemoryFlags, /* new flags */
+        Option<PID>, /* if present, indicates the process to modify */
     ),
 
     /// Pauses execution of the current thread and returns execution to the parent
     /// process.  This may return at any time in the future, including immediately.
+    ///
+    /// # Returns
+    ///
+    /// * **Ok**: The call completed successfully
+    ///
+    /// # Errors
+    ///
+    /// This syscall will never return an error.
     Yield,
 
     /// This process will now wait for an event such as an IRQ or Message.
+    ///
+    /// # Returns
+    ///
+    /// * **Ok**: The call completed successfully
+    ///
+    /// # Errors
+    ///
+    /// This syscall will never error.
     WaitEvent,
 
     /// This thread will now wait for a message with the given server ID. You
     /// can set up a pool by having multiple threads call `ReceiveMessage` with
     /// the same SID.
+    ///
+    /// # Returns
+    ///
+    /// * **Message**: A valid message from the queue
+    ///
+    /// # Errors
+    ///
+    /// * **ServerNotFound**: The given SID is not active or has terminated
+    /// * **ProcessNotFound**: The parent process terminated when we were getting ready
+    ///                        to block. This is an internal error.
+    /// * **BlockedProcess**: When running in Hosted mode, this indicates that this
+    ///                       thread is blocking.
     ReceiveMessage(SID),
 
     /// If a message is available for the specified server, return that message
     /// and resume execution. If no message is available, return `Result::None`
     /// immediately without blocking.
+    ///
+    /// # Returns
+    ///
+    /// * **Message**: A valid message from the queue
+    /// * **None**: Indicates that no message was in the queue
+    ///
+    /// # Errors
+    ///
+    /// * **ServerNotFound**: The given SID is not active or has terminated
+    /// * **ProcessNotFound**: The parent process terminated when we were getting ready
+    ///                        to block. This is an internal error.
     TryReceiveMessage(SID),
 
     /// Stop running the given process and return control to the parent. This
@@ -141,6 +197,10 @@ pub enum SysCall {
     /// will be called from within an interrupt context, but using the ordinary
     /// privilege level of the process.
     ///
+    /// # Returns
+    ///
+    /// * **Ok**: The interrupt has been mapped to this process
+    ///
     /// # Errors
     ///
     /// * **InterruptNotFound**: The specified interrupt isn't valid on this
@@ -210,6 +270,10 @@ pub enum SysCall {
     SwitchTo(PID, usize /* context ID */),
 
     /// Get a list of contexts that can be run in the given PID.
+    ///
+    /// # Errors
+    ///
+    /// * **UnhandledSyscall**: This syscall is currently unimplemented.
     ReadyThreads(PID),
 
     /// Create a new Server with a specified address
@@ -220,8 +284,8 @@ pub enum SysCall {
     ///
     /// # Returns
     ///
-    /// The specified SID, along with the connection ID for this process to talk
-    /// to the server.
+    /// * **NewServerID(sid, cid)**: The specified SID, along with the connection ID
+    ///                              for this process to talk to the server.
     ///
     /// # Errors
     ///
@@ -233,6 +297,10 @@ pub enum SysCall {
     /// Connect to a server.   This turns a 128-bit Server ID into a 32-bit
     /// Connection ID. Blocks until the server is available.
     ///
+    /// # Returns
+    ///
+    /// * **ConnectionID(cid)**: The new connection ID for communicating with the server.
+    ///
     /// # Errors
     ///
     /// None
@@ -241,15 +309,44 @@ pub enum SysCall {
     /// Try to connect to a server.   This turns a 128-bit Server ID into a 32-bit
     /// Connection ID.
     ///
+    /// # Returns
+    ///
+    /// * **ConnectionID(cid)**: The new connection ID for communicating with the server.
+    ///
     /// # Errors
     ///
     /// * **ServerNotFound**: The server could not be found.
     TryConnect(SID /* server id */),
 
     /// Send a message to a server (blocking until it's ready)
+    ///
+    /// # Returns
+    ///
+    /// * **Ok**: The Scalar / Send message was successfully sent, or the Borrow has finished
+    /// * **Scalar1**: The Server returned a `Scalar1` value
+    /// * **Scalar2**: The Server returned a `Scalar2` value
+    /// * **BlockedProcess**: In Hosted mode, the target process is now blocked
+    ///
+    /// # Errors
+    ///
+    /// * **ServerNotFound**: The server could not be found.
+    /// * **ProcessNotFound**: Internal error -- the parent process couldn't be found when blocking
     SendMessage(CID, Message),
 
     /// Try to send a message to a server
+    ///
+    /// # Returns
+    ///
+    /// * **Ok**: The Scalar / Send message was successfully sent, or the Borrow has finished
+    /// * **Scalar1**: The Server returned a `Scalar1` value
+    /// * **Scalar2**: The Server returned a `Scalar2` value
+    /// * **BlockedProcess**: In Hosted mode, the target process is now blocked
+    ///
+    /// # Errors
+    ///
+    /// * **ServerNotFound**: The server could not be found.
+    /// * **ServerQueueFull**: The server's mailbox is full
+    /// * **ProcessNotFound**: Internal error -- the parent process couldn't be found when blocking
     TrySendMessage(CID, Message),
 
     /// Return a Borrowed memory region to the sender
@@ -321,6 +418,13 @@ pub enum SysCall {
     /// Waits for a thread to finish, and returns the return value of that thread.
     JoinThread(TID),
 
+    /// A function to call when there is an exception such as a memory fault
+    /// or illegal instruction.
+    SetExceptionHandler(
+        usize, /* function pointer */
+        usize, /* stack pointer */
+    ),
+
     /// This syscall does not exist. It captures all possible
     /// arguments so detailed analysis can be performed.
     Invalid(usize, usize, usize, usize, usize, usize, usize),
@@ -363,6 +467,7 @@ pub enum SysCallNumber {
     DestroyServer = 34,
     Disconnect = 35,
     JoinThread = 36,
+    SetExceptionHandler = 37,
     Invalid,
 }
 
@@ -405,6 +510,7 @@ impl SysCallNumber {
             34 => DestroyServer,
             35 => Disconnect,
             36 => JoinThread,
+            37 => SetExceptionHandler,
             _ => Invalid,
         }
     }
@@ -551,10 +657,10 @@ impl SysCall {
             ],
             SysCall::UpdateMemoryFlags(a1, a2, a3) => [
                 SysCallNumber::UpdateMemoryFlags as usize,
-                a1.get(),
-                *a2 as usize,
-                crate::get_bits(a3),
-                0,
+                a1.as_mut_ptr() as usize,
+                a1.len(),
+                crate::get_bits(a2),
+                a3.map(|m| m.get() as usize).unwrap_or(0),
                 0,
                 0,
                 0,
@@ -736,6 +842,16 @@ impl SysCall {
                 0,
                 0,
             ],
+            SysCall::SetExceptionHandler(pc, sp) => [
+                SysCallNumber::SetExceptionHandler as usize,
+                *pc,
+                *sp,
+                0,
+                0,
+                0,
+                0,
+                0,
+            ],
             SysCall::Invalid(a1, a2, a3, a4, a5, a6, a7) => [
                 SysCallNumber::Invalid as usize,
                 *a1,
@@ -767,9 +883,9 @@ impl SysCall {
                 MemoryAddress::new(a3).ok_or(Error::InvalidSyscall)?,
                 crate::from_bits(a4).ok_or(Error::InvalidSyscall)?,
             ),
-            SysCallNumber::UnmapMemory => {
-                SysCall::UnmapMemory(MemoryRange::new(a1, a2).or(Err(Error::InvalidSyscall))?)
-            }
+            SysCallNumber::UnmapMemory => SysCall::UnmapMemory(unsafe {
+                MemoryRange::new(a1, a2).or(Err(Error::InvalidSyscall))
+            }?),
             SysCallNumber::Yield => SysCall::Yield,
             SysCallNumber::WaitEvent => SysCall::WaitEvent,
             SysCallNumber::ReceiveMessage => {
@@ -793,9 +909,9 @@ impl SysCall {
             ),
             SysCallNumber::DecreaseHeap => SysCall::DecreaseHeap(a1 as usize),
             SysCallNumber::UpdateMemoryFlags => SysCall::UpdateMemoryFlags(
-                MemoryAddress::new(a1).ok_or(Error::InvalidSyscall)?,
-                a2 as usize,
+                unsafe { MemoryRange::new(a1, a2) }?,
                 crate::from_bits(a3).ok_or(Error::InvalidSyscall)?,
+                PID::new(a4 as _),
             ),
             SysCallNumber::SetMemRegion => SysCall::SetMemRegion(
                 pid_from_usize(a1)?,
@@ -815,7 +931,7 @@ impl SysCall {
                 .unwrap_or_else(|_| SysCall::Invalid(a1, a2, a3, a4, a5, a6, a7)),
             SysCallNumber::ReturnMemory => SysCall::ReturnMemory(
                 MessageSender::from_usize(a1),
-                MemoryRange::new(a2, a3)?,
+                unsafe { MemoryRange::new(a2, a3) }?,
                 MemorySize::new(a4),
                 MemorySize::new(a5),
             ),
@@ -835,7 +951,7 @@ impl SysCall {
                     a1 as u32,
                     Message::MutableBorrow(MemoryMessage {
                         id: a3,
-                        buf: MemoryRange::new(a4, a5)?,
+                        buf: unsafe { MemoryRange::new(a4, a5) }?,
                         offset: MemoryAddress::new(a6),
                         valid: MemorySize::new(a7),
                     }),
@@ -844,7 +960,7 @@ impl SysCall {
                     a1 as u32,
                     Message::Borrow(MemoryMessage {
                         id: a3,
-                        buf: MemoryRange::new(a4, a5)?,
+                        buf: unsafe { MemoryRange::new(a4, a5) }?,
                         offset: MemoryAddress::new(a6),
                         valid: MemorySize::new(a7),
                     }),
@@ -853,7 +969,7 @@ impl SysCall {
                     a1 as u32,
                     Message::Move(MemoryMessage {
                         id: a3,
-                        buf: MemoryRange::new(a4, a5)?,
+                        buf: unsafe { MemoryRange::new(a4, a5) }?,
                         offset: MemoryAddress::new(a6),
                         valid: MemorySize::new(a7),
                     }),
@@ -898,6 +1014,7 @@ impl SysCall {
             }
             SysCallNumber::Disconnect => SysCall::Disconnect(a1 as _),
             SysCallNumber::JoinThread => SysCall::JoinThread(a1 as _),
+            SysCallNumber::SetExceptionHandler => SysCall::SetExceptionHandler(a1 as _, a2 as _),
             SysCallNumber::Invalid => SysCall::Invalid(a1, a2, a3, a4, a5, a6, a7),
         })
     }
@@ -965,12 +1082,33 @@ impl SysCall {
         }
     }
 
+    /// If the syscall has memory attached to it, return the memory, mutably
+    ///
+    /// # Safety
+    ///
+    /// This function is only safe to call to fixup the pointer. It should
+    /// not be used for any other purpose.
+    pub unsafe fn memory_mut(&mut self) -> Option<&mut MemoryRange> {
+        match self {
+            SysCall::TrySendMessage(_, msg) | SysCall::SendMessage(_, msg) => match msg {
+                Message::Move(memory_message)
+                | Message::Borrow(memory_message)
+                | Message::MutableBorrow(memory_message) => Some(&mut memory_message.buf),
+                _ => None,
+            },
+            SysCall::ReturnMemory(_, range, _, _) => Some(range),
+            _ => None,
+        }
+    }
+
     /// Returns `true` if the given syscall may be called from an IRQ context
     pub fn can_call_from_interrupt(&self) -> bool {
+        if let SysCall::TrySendMessage(_cid, msg) = self {
+            return !msg.is_blocking();
+        }
         matches!(
             self,
-            SysCall::TrySendMessage(_, _)
-                | SysCall::TryConnect(_)
+            SysCall::TryConnect(_)
                 | SysCall::TryReceiveMessage(_)
                 | SysCall::ReturnToParent(_, _)
                 | SysCall::ReturnScalar2(_, _, _)
@@ -980,31 +1118,6 @@ impl SysCall {
     }
 }
 
-extern "Rust" {
-    // fn _xous_syscall_rust(
-    //     nr: usize,
-    //     a1: usize,
-    //     a2: usize,
-    //     a3: usize,
-    //     a4: usize,
-    //     a5: usize,
-    //     a6: usize,
-    //     a7: usize,
-    //     ret: &mut Result,
-    // );
-    fn _xous_syscall(
-        nr: usize,
-        a1: usize,
-        a2: usize,
-        a3: usize,
-        a4: usize,
-        a5: usize,
-        a6: usize,
-        a7: usize,
-        ret: &mut Result,
-    );
-}
-
 /// Map the given physical address to the given virtual address.
 /// The `size` field must be page-aligned.
 pub fn map_memory(
@@ -1046,6 +1159,22 @@ pub fn unmap_memory(range: MemoryRange) -> core::result::Result<(), Error> {
     }
 }
 
+/// Update the permissions on the given memory range. Note that permissions may
+/// only be stripped here -- they may never be added.
+pub fn update_memory_flags(
+    range: MemoryRange,
+    flags: MemoryFlags,
+) -> core::result::Result<Result, Error> {
+    let result = rsyscall(SysCall::UpdateMemoryFlags(range, flags, None))?;
+    if let Result::Ok = result {
+        Ok(Result::Ok)
+    } else if let Result::Error(e) = result {
+        Err(e)
+    } else {
+        Err(Error::InternalError)
+    }
+}
+
 /// Map the given physical address to the given virtual address.
 /// The `size` field must be page-aligned.
 pub fn return_memory(sender: MessageSender, mem: MemoryRange) -> core::result::Result<(), Error> {
@@ -1360,16 +1489,16 @@ pub fn terminate_process(exit_code: u32) -> ! {
 /// Return execution to the kernel. This function may return at any time,
 /// including immediately
 pub fn yield_slice() {
-    rsyscall(SysCall::Yield).expect("yield_slice returned an error");
+    rsyscall(SysCall::Yield).ok();
 }
 
 /// Return execution to the kernel and wait for a message or an interrupt.
 pub fn wait_event() {
-    rsyscall(SysCall::WaitEvent).expect("wait_event returned an error");
+    rsyscall(SysCall::WaitEvent).ok();
 }
 
 #[deprecated(
-    since = "0.2",
+    since = "0.2.0",
     note = "Please use create_thread_n() or create_thread()"
 )]
 pub fn create_thread_simple<T, U>(
@@ -1501,7 +1630,7 @@ pub fn wait_thread<T>(joiner: crate::arch::WaitHandle<T>) -> SysCallResult {
 }
 
 /// Create a new process by running it in its own thread
-#[cfg(not(any(target_os = "none", target_os = "xous")))]
+#[cfg(feature = "processes-as-threads")]
 pub fn create_process_as_thread<F>(
     args: ProcessArgsAsThread<F>,
 ) -> core::result::Result<crate::arch::ProcessHandleAsThread, Error>
@@ -1519,7 +1648,7 @@ where
 }
 
 /// Wait for a thread to finish
-#[cfg(not(any(target_os = "none", target_os = "xous")))]
+#[cfg(feature = "processes-as-threads")]
 pub fn wait_process_as_thread(joiner: crate::arch::ProcessHandleAsThread) -> SysCallResult {
     crate::arch::wait_process_as_thread(joiner)
 }
@@ -1577,6 +1706,13 @@ pub fn destroy_server(sid: SID) -> core::result::Result<(), Error> {
 /// Disconnect the specified connection ID and mark it as free. This
 /// connection ID may be reused by the server in the future, so ensure
 /// no other threads are using the connection ID before disposing of it.
+///
+/// # Safety
+///
+/// This function must only be called when the connection is no longer in
+/// use. Calling this function when the connection ID is in use will result
+/// in kernel errors or, if the CID is reused, silent failures due to
+/// messages going to the wrong server.
 pub unsafe fn disconnect(cid: CID) -> core::result::Result<(), Error> {
     rsyscall(SysCall::Disconnect(cid)).and_then(|result| {
         if let Result::Ok = result {
@@ -1604,21 +1740,47 @@ pub fn join_thread(tid: TID) -> core::result::Result<usize, Error> {
         }
     })
 }
+/* https://github.com/betrusted-io/xous-core/issues/90
+static EXCEPTION_HANDLER: core::sync::atomic::AtomicUsize = core::sync::atomic::AtomicUsize::new(0);
+fn handle_exception(exception_type: usize, arg1: usize, arg2: usize) -> isize {
+    let exception = crate::exceptions::Exception::new(exception_type, arg1, arg2);
+    let f = EXCEPTION_HANDLER.load(core::sync::atomic::Ordering::SeqCst);
+    let f = unsafe { core::mem::transmute::<usize, fn(Exception) -> isize>(f) };
+    f(exception)
+}
+/// Sets the given function as this process' Exception handler. This function
+/// will be called whenever an Exception occurs such as a memory fault,
+/// illegal instruction, or a child process terminating.
+pub fn set_exception_handler(
+    handler: fn(crate::Exception) -> isize,
+) -> core::result::Result<(), Error> {
+    #[cfg(feature = "bit-flags")]
+    let flags = crate::MemoryFlags::R | crate::MemoryFlags::W | crate::MemoryFlags::RESERVE;
+    #[cfg(not(feature = "bit-flags"))]
+    let flags = 0b0000_0010 | 0b0000_0100 | 0b0000_0001;
+
+    let stack = crate::map_memory(None, None, 131_072, flags)?;
+    EXCEPTION_HANDLER.store(handler as usize, core::sync::atomic::Ordering::SeqCst);
+    rsyscall(SysCall::SetExceptionHandler(
+        handle_exception as usize,
+        stack.as_ptr() as usize,
+    ))
+    .and_then(|result| {
+        if let Result::Ok = result {
+            Ok(())
+        } else if let Result::Error(Error::ThreadNotAvailable) = result {
+            Err(Error::ThreadNotAvailable)
+        } else {
+            Err(Error::InternalError)
+        }
+    })
+}
+*/
 
 /// Perform a raw syscall and return the result. This will transform
 /// `xous::Result::Error(e)` into an `Err(e)`.
 pub fn rsyscall(call: SysCall) -> SysCallResult {
-    let mut ret = Result::Ok;
-    let args = call.as_args();
-    unsafe {
-        _xous_syscall(
-            args[0], args[1], args[2], args[3], args[4], args[5], args[6], args[7], &mut ret,
-        )
-    };
-    match ret {
-        Result::Error(e) => Err(e),
-        other => Ok(other),
-    }
+    crate::arch::syscall(call)
 }
 
 // /// This is dangerous, but fast.
